@@ -1,5 +1,5 @@
 /* libdiskfs implementation of fs.defs: dir_link
-   Copyright (C) 1992, 1993, 1994, 1995, 1996, 1997 Free Software Foundation
+   Copyright (C) 1992,93,94,95,96,97,99 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -31,26 +31,26 @@ diskfs_S_dir_link (struct protid *dircred,
   struct dirstat *ds = alloca (diskfs_dirstat_size);
   error_t error;
 
-  if (!filecred)
+  if (!dircred)
     return EOPNOTSUPP;
-  
-  np = filecred->po->np;
+
   if (diskfs_check_readonly ())
     return EROFS;
-  
-  if (!dircred)
+
+  if (!filecred)
     return EXDEV;
-  
-  mutex_lock (&np->lock);
+
+  np = filecred->po->np;
+  pthread_mutex_lock (&np->lock);
   if (S_ISDIR (np->dn_stat.st_mode))
     {
-      mutex_unlock (&np->lock);
-      return EISDIR;
+      pthread_mutex_unlock (&np->lock);
+      return EPERM;
     }
-  mutex_unlock (&np->lock);
-  
+  pthread_mutex_unlock (&np->lock);
+
   dnp = dircred->po->np;
-  mutex_lock (&dnp->lock);
+  pthread_mutex_lock (&dnp->lock);
 
   /* Lookup new location */
   error = diskfs_lookup (dnp, name, RENAME, &tnp, ds, dircred);
@@ -64,45 +64,45 @@ diskfs_S_dir_link (struct protid *dircred,
       if (error == EAGAIN)
 	error = EINVAL;
       diskfs_drop_dirstat (dnp, ds);
-      mutex_unlock (&dnp->lock);
+      pthread_mutex_unlock (&dnp->lock);
       return error;
     }
 
   if (np == tnp)
     {
       diskfs_drop_dirstat (dnp, ds);
-      mutex_unlock (&dnp->lock);
-      mutex_unlock (&tnp->lock);
+      pthread_mutex_unlock (&dnp->lock);
+      pthread_mutex_unlock (&tnp->lock);
       mach_port_deallocate (mach_task_self (), filecred->pi.port_right);
       return 0;
     }
-  
+
   if (tnp && S_ISDIR (tnp->dn_stat.st_mode))
     {
       diskfs_drop_dirstat (dnp, ds);
-      mutex_unlock (&dnp->lock);
-      mutex_unlock (&tnp->lock);
+      pthread_mutex_unlock (&dnp->lock);
+      pthread_mutex_unlock (&tnp->lock);
       return EISDIR;
     }
-  
+
   /* Create new entry for NP */
 
   /* This is safe because NP is not a directory (thus not DNP) and
      not TNP and is a leaf. */
-  mutex_lock (&np->lock);
-  
+  pthread_mutex_lock (&np->lock);
+
   /* Increment link count */
   if (np->dn_stat.st_nlink == diskfs_link_max - 1)
     {
       diskfs_drop_dirstat (dnp, ds);
-      mutex_unlock (&np->lock);
-      mutex_unlock (&dnp->lock);
+      pthread_mutex_unlock (&np->lock);
+      pthread_mutex_unlock (&dnp->lock);
       return EMLINK;
     }
   np->dn_stat.st_nlink++;
   np->dn_set_ctime = 1;
-  diskfs_node_update (np, 1);
-  
+  diskfs_node_update (np, diskfs_synchronous);
+
   /* Attach it */
   if (tnp)
     {
@@ -120,12 +120,12 @@ diskfs_S_dir_link (struct protid *dircred,
     }
   else
     error = diskfs_direnter (dnp, name, np, ds, dircred);
-  
+
   if (diskfs_synchronous)
     diskfs_node_update (dnp, 1);
-  
-  mutex_unlock (&dnp->lock);
-  mutex_unlock (&np->lock);
+
+  pthread_mutex_unlock (&dnp->lock);
+  pthread_mutex_unlock (&np->lock);
   if (!error)
     /* MiG won't do this for us, which it ought to. */
     mach_port_deallocate (mach_task_self (), filecred->pi.port_right);
