@@ -1,5 +1,5 @@
 /* Map the disk image and handle faults accessing it.
-   Copyright (C) 1996, 1997 Free Software Foundation, Inc.
+   Copyright (C) 1996,97,99,2001,02 Free Software Foundation, Inc.
    Written by Roland McGrath.
 
    This program is free software; you can redistribute it and/or
@@ -21,6 +21,8 @@
 #include <hurd/sigpreempt.h>
 #include <error.h>
 
+__thread struct disk_image_user *diskfs_exception_diu;
+
 struct pager *diskfs_disk_pager;
 
 static void fault_handler (int sig, long int sigcode, struct sigcontext *scp);
@@ -33,32 +35,42 @@ static struct hurd_signal_preemptor preemptor =
 
 /* A top-level function for the paging thread that just services paging
    requests.  */
-static void
-service_paging_requests (any_t arg)
+static void *
+service_paging_requests (void *arg)
 {
   struct port_bucket *pager_bucket = arg;
   for (;;)
     ports_manage_port_operations_multithread (pager_bucket,
 					      pager_demuxer,
-					      1000 * 60 * 2, 1000 * 60 * 10,
-					      1, MACH_PORT_NULL);
+					      1000 * 60 * 2,
+					      1000 * 60 * 10, 0);
+  return NULL;
 }
 
 void
 diskfs_start_disk_pager (struct user_pager_info *upi,
-			 struct port_bucket *pager_bucket, int may_cache,
+			 struct port_bucket *pager_bucket,
+			 int may_cache, int notify_on_evict,
 			 size_t size, void **image)
 {
+  pthread_t thread;
   error_t err;
   mach_port_t disk_pager_port;
 
   /* Make a thread to service paging requests.  */
-  cthread_detach (cthread_fork ((cthread_fn_t) service_paging_requests,
-				(any_t)pager_bucket));
+  err = pthread_create (&thread, NULL, service_paging_requests, pager_bucket);
+  if (!err)
+    pthread_detach (thread);
+  else
+    {
+      errno = err;
+      perror ("pthread_create");
+    }
 
   /* Create the pager.  */
   diskfs_disk_pager = pager_create (upi, pager_bucket,
-				    may_cache, MEMORY_OBJECT_COPY_NONE);
+				    may_cache, MEMORY_OBJECT_COPY_NONE,
+				    notify_on_evict);
   assert (diskfs_disk_pager);
 
   /* Get a port to the disk pager.  */
@@ -87,13 +99,26 @@ diskfs_start_disk_pager (struct user_pager_info *upi,
 static void
 fault_handler (int sig, long int sigcode, struct sigcontext *scp)
 {
-  jmp_buf *env = cthread_data (cthread_self ());
   error_t err;
 
-  assert (env && "unexpected fault on disk image");
+#ifndef NDEBUG
+  if (diskfs_exception_diu == NULL)
+    {
+      error (0, 0,
+	     "BUG: unexpected fault on disk image (%d, %#lx) in [%#lx,%#lx)"
+	     " eip %#zx err %#x",
+	     sig, sigcode,
+	     preemptor.first, preemptor.last,
+	     scp->sc_pc, scp->sc_error);
+      assert (scp->sc_error == EKERN_MEMORY_ERROR);
+      err = pager_get_error (diskfs_disk_pager, sigcode);
+      assert (err);
+      assert_perror (err);
+    }
+#endif
 
   /* Clear the record, since the faulting thread will not.  */
-  cthread_set_data (cthread_self (), 0);
+  diskfs_exception_diu = NULL;
 
   /* Fetch the error code from the pager.  */
   assert (scp->sc_error == EKERN_MEMORY_ERROR);
@@ -101,5 +126,5 @@ fault_handler (int sig, long int sigcode, struct sigcontext *scp)
   assert (err);
 
   /* Make `diskfault_catch' return the error code.  */
-  longjmp (*env, err);
+  longjmp (diskfs_exception_diu->env, err);
 }
