@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 1994, 1995, 1996 Free Software Foundation
+   Copyright (C) 1994,95,96,2000,01 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -26,36 +26,41 @@ diskfs_S_io_reauthenticate (struct protid *cred,
   struct protid *newcred;
   error_t err;
   mach_port_t newright;
+  struct iouser *user;
 
   if (cred == 0)
     return EOPNOTSUPP;
 
-  /* This routine must carefully ignore EINTR because we 
+  /* This routine must carefully ignore EINTR because we
      are a simpleroutine, so callers won't know to restart. */
 
-  mutex_lock (&cred->po->np->lock);
+  pthread_mutex_lock (&cred->po->np->lock);
   do
     err = diskfs_start_protid (cred->po, &newcred);
   while (err == EINTR);
   if (err)
     {
-      mutex_unlock (&cred->po->np->lock);
+      pthread_mutex_unlock (&cred->po->np->lock);
       return err;
     }
 
-  newright = ports_get_right (newcred);
-  err = mach_port_insert_right (mach_task_self (), newright, newright,
-				MACH_MSG_TYPE_MAKE_SEND);
-  assert_perror (err);
+  newright = ports_get_send_right (newcred);
+  assert (newright != MACH_PORT_NULL);
+
+  err = iohelp_reauth (&user, diskfs_auth_server_port, rend_port,
+		       newright, 1);
+  if (! err)
+    {
+      diskfs_finish_protid (newcred, user);
+      iohelp_free_iouser (user);
+      mach_port_deallocate (mach_task_self (), rend_port);
+    }
 
-  diskfs_finish_protid (newcred, iohelp_reauth (diskfs_auth_server_port, 
-						rend_port, newright, 1));
-  mach_port_deallocate (mach_task_self (), rend_port);
   mach_port_deallocate (mach_task_self (), newright);
-    
-  mutex_unlock (&cred->po->np->lock);
+
+  pthread_mutex_unlock (&cred->po->np->lock);
 
   ports_port_deref (newcred);
 
-  return 0;
+  return err;
 }
