@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 1994, 1995, 1996, 1997 Free Software Foundation, Inc.
+   Copyright (C) 1994,95,96,97,2001 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -29,7 +29,7 @@ diskfs_S_io_write (struct protid *cred,
 {
   struct node *np;
   error_t err;
-  int off = offset;
+  off_t off = offset;
 
   if (!cred)
     return EOPNOTSUPP;
@@ -38,7 +38,7 @@ diskfs_S_io_write (struct protid *cred,
   if (!(cred->po->openstat & O_WRITE))
     return EBADF;
 
-  mutex_lock (&np->lock);
+  pthread_mutex_lock (&np->lock);
 
   assert (!S_ISDIR(np->dn_stat.st_mode));
 
@@ -64,6 +64,8 @@ diskfs_S_io_write (struct protid *cred,
 	diskfs_node_update (np, 1);
       if (err)
 	goto out;
+      if (np->filemod_reqs)
+	diskfs_notice_filechange (np, FILE_CHANGED_EXTEND, 0, off + datalen);
     }
 
   if (off + (off_t) datalen > np->dn_stat.st_size)
@@ -84,7 +86,9 @@ diskfs_S_io_write (struct protid *cred,
       && ((cred->po->openstat & O_FSYNC) || diskfs_synchronous))
     diskfs_file_update (np, 1);
 
+  if (!err && np->filemod_reqs)
+    diskfs_notice_filechange (np, FILE_CHANGED_WRITE, off, off + *amt);
  out:
-  mutex_unlock (&np->lock);
+  pthread_mutex_unlock (&np->lock);
   return err;
 }
