@@ -1,8 +1,9 @@
 /* Standard startup-time command line parser
 
-   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
+   Copyright (C) 1995, 1996, 1997, 1998, 1999, 2001, 2007
+     Free Software Foundation, Inc.
 
-   Written by Miles Bader <miles@gnu.ai.mit.edu>
+   Written by Miles Bader <miles@gnu.org>
 
    This file is part of the GNU Hurd.
 
@@ -23,9 +24,13 @@
 #include <stdio.h>
 #include <argp.h>
 #include <hurd/store.h>
+#include <hurd/paths.h>
 #include "priv.h"
 
-char *diskfs_boot_flags = 0;
+const char *diskfs_boot_command_line;
+char **_diskfs_boot_command;
+
+int _diskfs_boot_pause;
 
 extern char **diskfs_argv;
 
@@ -36,16 +41,33 @@ mach_port_t diskfs_exec_server_task = MACH_PORT_NULL;
 #define OPT_HOST_PRIV_PORT	(-1)
 #define OPT_DEVICE_MASTER_PORT	(-2)
 #define OPT_EXEC_SERVER_TASK	(-3)
-#define OPT_BOOTFLAGS		(-4)
+#define OPT_BOOT_CMDLINE	(-4)
+#define OPT_BOOT_COMMAND	(-5)
+#define OPT_BOOT_INIT_PROGRAM	(-6)
+#define OPT_BOOT_PAUSE		(-7)
 
 static const struct argp_option
 startup_options[] =
 {
+  {"directory",		 'C',			 "DIRECTORY", 0,
+   "Use DIRECTORY as the root of the filesystem"},
+  {"virtual-root",	 0, 0, OPTION_ALIAS},
+  {"chroot",		 0, 0, OPTION_ALIAS},
+
   {0,0,0,0, "Boot options:", -2},
+  {"multiboot-command-line", OPT_BOOT_CMDLINE, "ARGS", 0,
+   "Required for bootstrap filesystem, the multiboot kernel command line"},
+  {"bootflags", 0, 0, OPTION_ALIAS|OPTION_HIDDEN},
+  {"boot-init-program",  OPT_BOOT_INIT_PROGRAM,  "FILE", 0,
+   "For bootstrap filesystem, init program to run (default " _HURD_INIT ")"},
+  {"boot-debug-pause",  OPT_BOOT_PAUSE,	         0, 0,
+   "Pause for keystroke before starting bootstrap programs"},
+  {"boot-command",	 OPT_BOOT_COMMAND,	 0, 0,
+   "Remaining arguments form command line to run"
+   " at bootstrap instead of init"},
   {"host-priv-port",     OPT_HOST_PRIV_PORT,     "PORT"},
   {"device-master-port", OPT_DEVICE_MASTER_PORT, "PORT"},
   {"exec-server-task",   OPT_EXEC_SERVER_TASK,   "PORT"},
-  {"bootflags",          OPT_BOOTFLAGS,          "FLAGS"},
 
   {0}
 };
@@ -55,10 +77,17 @@ parse_startup_opt (int opt, char *arg, struct argp_state *state)
 {
   switch (opt)
     {
-    case 'r':
-      diskfs_readonly = 1; break;
-    case 'w':
-      diskfs_readonly = 0; break;
+#define TOGGLE(var, on, off) \
+    case on: var = 1; break; \
+    case off: var = 0; break;
+      TOGGLE (diskfs_readonly, 'r', 'w');
+      TOGGLE (_diskfs_nosuid, 'S', OPT_SUID_OK);
+      TOGGLE (_diskfs_noexec, 'E', OPT_EXEC_OK);
+      TOGGLE (_diskfs_noatime, 'A', OPT_ATIME);
+      TOGGLE (_diskfs_no_inherit_dir_group, OPT_NO_INHERIT_DIR_GROUP,
+	      OPT_INHERIT_DIR_GROUP);
+#undef	TOGGLE
+
     case 's':
       if (arg == NULL)
 	diskfs_synchronous = 1;
@@ -77,8 +106,22 @@ parse_startup_opt (int opt, char *arg, struct argp_state *state)
       _hurd_host_priv = atoi (arg); break;
     case OPT_EXEC_SERVER_TASK:
       diskfs_exec_server_task = atoi (arg); break;
-    case OPT_BOOTFLAGS:
-      diskfs_boot_flags = arg; break;
+    case OPT_BOOT_CMDLINE:
+      diskfs_boot_command_line = arg; break;
+    case OPT_BOOT_INIT_PROGRAM:
+      diskfs_boot_init_program = arg; break;
+    case OPT_BOOT_PAUSE:
+      _diskfs_boot_pause = 1; break;
+    case 'C':
+      _diskfs_chroot_directory = arg; break;
+
+    case OPT_BOOT_COMMAND:
+      if (state->next == state->argc)
+	argp_error (state, "Command line must follow --boot-command option");
+      _diskfs_boot_command = state->argv + state->next;
+      state->next = state->argc; /* stop parsing */
+      {char **p; for (p = _diskfs_boot_command; *p; ++p) printf("BC %s\n",*p);}
+      break;
 
     case ARGP_KEY_END:
       diskfs_argv = state->argv; break;
