@@ -22,11 +22,11 @@
 void
 diskfs_release_peropen (struct peropen *po)
 {
-  mutex_lock (&po->np->lock);
+  pthread_mutex_lock (&po->np->lock);
 
   if (--po->refcnt)
     {
-      mutex_unlock (&po->np->lock);
+      pthread_mutex_unlock (&po->np->lock);
       return;
     }
 
@@ -34,10 +34,8 @@ diskfs_release_peropen (struct peropen *po)
     mach_port_deallocate (mach_task_self (), po->root_parent);
 
   if (po->shadow_root && po->shadow_root != po->np)
-    {
-      mutex_lock (&po->shadow_root->lock);
-      diskfs_nput (po->shadow_root);
-    }
+    diskfs_nrele (po->shadow_root);
+
   if (po->shadow_root_parent)
     mach_port_deallocate (mach_task_self (), po->shadow_root_parent);
 
@@ -47,5 +45,6 @@ diskfs_release_peropen (struct peropen *po)
 
   diskfs_nput (po->np);
 
+  free (po->path);
   free (po);
 }
