@@ -1,6 +1,6 @@
 /* Unimplemented rpcs from <hurd/fs.defs>
 
-   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
+   Copyright (C) 1995, 1996, 1999 Free Software Foundation, Inc.
    Written by Michael I. Bushnell, p/BSG.
 
    This file is part of the GNU Hurd.
@@ -29,14 +29,6 @@ netfs_S_file_notice_changes (struct protid *user,
   return EOPNOTSUPP;
 }
 
-error_t
-netfs_S_file_get_translator_cntl (struct protid *user,
-				  mach_port_t *trans,
-				  mach_msg_type_name_t *transtype)
-{
-  return EOPNOTSUPP;
-}
-
 error_t
 netfs_S_file_getfh (struct protid *user,
 		    char **data, mach_msg_type_number_t *ndata)
@@ -50,21 +42,3 @@ netfs_S_ifsock_getsockaddr (struct protid *user,
 {
   return EOPNOTSUPP;
 }
-
-error_t
-netfs_S_io_pathconf (struct protid *user,
-		     int name,
-		     int *value)
-{
-  return EOPNOTSUPP;
-}
-
-error_t
-netfs_S_io_server_version (struct protid *user,
-			   char *name,
-			   int *major,
-			   int *minor,
-			   int *edit)
-{
-  return EOPNOTSUPP;
-}
