@@ -1,5 +1,5 @@
 /* 
-   Copyright (C) 1996 Free Software Foundation, Inc.
+   Copyright (C) 1996, 2001 Free Software Foundation, Inc.
    Written by Michael I. Bushnell, p/BSG.
 
    This file is part of the GNU Hurd.
@@ -23,16 +23,17 @@
 
 error_t
 netfs_S_fsys_syncfs (mach_port_t cntl,
+		     mach_port_t reply,
+		     mach_msg_type_name_t reply_type,
 		     int wait,
 		     int children)
 {
   struct iouser *cred;
-  uid_t root = 0;
   error_t err;
 
-  cred = iohelp_create_iouser (make_idvec (), make_idvec ());
-  idvec_set_ids (cred->uids, &root, 1);
-  idvec_set_ids (cred->gids, &root, 1);
+  err = iohelp_create_simple_iouser (&cred, 0, 0);
+  if (err)
+    return err;
   err = netfs_attempt_syncfs (cred, wait);
   iohelp_free_iouser (cred);
   return err;
