@@ -1,5 +1,5 @@
 /* 
-   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
+   Copyright (C) 1995,96,2001 Free Software Foundation, Inc.
    Written by Michael I. Bushnell, p/BSG.
 
    This file is part of the GNU Hurd.
@@ -26,13 +26,18 @@ netfs_S_io_duplicate (struct protid *user,
 		      mach_port_t *newport,
 		      mach_msg_type_name_t *newporttp)
 {
+  error_t err;
   struct protid *newpi;
+  struct iouser *clone;
+
+  err = iohelp_dup_iouser (&clone, user->user);
+  if (err)
+    return err;
   
-  mutex_lock (&user->po->np->lock);
-  newpi = netfs_make_protid (user->po,
-			     iohelp_dup_iouser (user->user));
+  pthread_mutex_lock (&user->po->np->lock);
+  newpi = netfs_make_protid (user->po, clone);
   *newport = ports_get_right (newpi);
-  mutex_unlock (&user->po->np->lock);
+  pthread_mutex_unlock (&user->po->np->lock);
   *newporttp = MACH_MSG_TYPE_MAKE_SEND;
   ports_port_deref (newpi);
   return 0;
