@@ -1,5 +1,5 @@
-/* 
-   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
+/*
+   Copyright (C) 1995,96,2001,02 Free Software Foundation, Inc.
    Written by Michael I. Bushnell, p/BSG.
 
    This file is part of the GNU Hurd.
@@ -21,16 +21,6 @@
 #include "netfs.h"
 #include "io_S.h"
 
-static inline int
-listmember (int *list, int query, int n)
-{
-  int i;
-  for (i = 0; i < n; i++)
-    if (list[i] == query)
-      return 1;
-  return 0;
-}
-
 error_t
 netfs_S_io_restrict_auth (struct protid *user,
 			  mach_port_t *newport,
@@ -40,39 +30,33 @@ netfs_S_io_restrict_auth (struct protid *user,
 			  gid_t *gids,
 			  mach_msg_type_number_t ngids)
 {
-  struct idvec *uvec, *gvec;
-  int i;
+  error_t err;
   struct protid *newpi;
-  
+  struct iouser *new_user;
+
   if (!user)
     return EOPNOTSUPP;
-  
-  uvec = make_idvec ();
-  gvec = make_idvec ();
 
-  if (idvec_contains (user->user->uids, 0))
+  err = iohelp_restrict_iouser (&new_user, user->user,
+				uids, nuids, gids, ngids);
+  if (err)
+    return err;
+
+  pthread_mutex_lock (&user->po->np->lock);
+  newpi = netfs_make_protid (user->po, new_user);
+  if (newpi)
     {
-      idvec_set_ids (uvec, uids, nuids);
-      idvec_set_ids (gvec, gids, ngids);
+      *newport = ports_get_right (newpi);
+      pthread_mutex_unlock (&user->po->np->lock);
+      *newporttype = MACH_MSG_TYPE_MAKE_SEND;
     }
   else
     {
-      for (i = 0; i < user->user->uids->num; i++)
-	if (listmember (uids, user->user->uids->ids[i], nuids))
-	  idvec_add (uvec, user->user->uids->ids[i]);
-      
-      for (i = 0; i < user->user->gids->num; i++)
-	if (listmember (gids, user->user->gids->ids[i], ngids))
-	  idvec_add (gvec, user->user->gids->ids[i]);
+      pthread_mutex_unlock (&user->po->np->lock);
+      iohelp_free_iouser (new_user);
+      err = ENOMEM;
     }
-  
-  mutex_lock (&user->po->np->lock);
-  newpi = netfs_make_protid (user->po, iohelp_create_iouser (uvec, gvec));
-  *newport = ports_get_right (newpi);
-  mutex_unlock (&user->po->np->lock);
-  
-  *newporttype = MACH_MSG_TYPE_MAKE_SEND;
-  ports_port_deref (newpi);
-  return 0;
-}
 
+  ports_port_deref (newpi);
+  return err;
+}
