@@ -1,5 +1,5 @@
 /* 
-   Copyright (C) 1996 Free Software Foundation, Inc.
+   Copyright (C) 1996, 1997 Free Software Foundation, Inc.
    Written by Michael I. Bushnell, p/BSG.
 
    This file is part of the GNU Hurd.
@@ -23,13 +23,25 @@
 void
 netfs_release_peropen (struct peropen *po)
 {
-  mutex_lock (&po->np->lock);
+  pthread_mutex_lock (&po->np->lock);
   if (--po->refcnt)
-    mutex_unlock (&po->np->lock);
+    pthread_mutex_unlock (&po->np->lock);
   else
     {
-      mach_port_deallocate (mach_task_self (), po->dotdotport);
+      if (po->root_parent)
+	mach_port_deallocate (mach_task_self (), po->root_parent);
+
+      if (po->shadow_root && po->shadow_root != po->np)
+	{
+	  pthread_mutex_lock (&po->shadow_root->lock);
+	  netfs_nput (po->shadow_root);
+	}
+      if (po->shadow_root_parent)
+	mach_port_deallocate (mach_task_self (), po->shadow_root_parent);
+
       netfs_nput (po->np);
+
+      free (po->path);
       free (po);
     }
 }
