@@ -1,5 +1,5 @@
 /* Wrapper for unsolicited memory_object_data_supply
-   Copyright (C) 1996, 1997 Free Software Foundation, Inc.
+   Copyright (C) 1996, 1997, 2000 Free Software Foundation, Inc.
    Written by Michael I. Bushnell, p/BSG.
 
    This file is part of the GNU Hurd.
@@ -28,17 +28,17 @@ pager_offer_page (struct pager *p,
 		  vm_offset_t offset,
 		  vm_address_t buf)
 {
-  mutex_lock (&p->interlock);
+  pthread_mutex_lock (&p->interlock);
 
   if (_pager_pagemap_resize (p, offset + vm_page_size))
     {
-      char *pm_entry = &p->pagemap[offset / vm_page_size];
+      short *pm_entry = &p->pagemap[offset / vm_page_size];
 
       while (*pm_entry & PM_INCORE)
 	{
-	  mutex_unlock (&p->interlock);
+	  pthread_mutex_unlock (&p->interlock);
 	  pager_flush_some (p, offset, vm_page_size, 1);
-	  mutex_lock (&p->interlock);
+	  pthread_mutex_lock (&p->interlock);
 	}
       *pm_entry |= PM_INCORE;
 
@@ -47,6 +47,6 @@ pager_offer_page (struct pager *p,
 				 precious, MACH_PORT_NULL);
     }
 
-  mutex_unlock (&p->interlock);
+  pthread_mutex_unlock (&p->interlock);
 }
 
