@@ -1,5 +1,5 @@
-/* 
-   Copyright (C) 1994 Free Software Foundation
+/*
+   Copyright (C) 1994,2002 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -16,12 +16,12 @@
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */
 
 #include "priv.h"
-#include "fs_S.h"
 
 kern_return_t
 trivfs_S_file_get_translator (struct trivfs_protid *cred,
-			      mach_port_t reply, mach_msg_type_name_t reply_type,
-			      char **trans, u_int *translen)
+			      mach_port_t reply,
+			      mach_msg_type_name_t reply_type,
+			      char **trans, size_t *translen)
 {
   return EOPNOTSUPP;
 }
