@@ -1,6 +1,6 @@
 /* Reparent a directory
 
-   Copyright (C) 1997 Free Software Foundation
+   Copyright (C) 1997,2002,2010 Free Software Foundation, Inc.
 
    This file is part of the GNU Hurd.
 
@@ -9,7 +9,7 @@
    the Free Software Foundation; either version 2, or (at your option)
    any later version.
 
-   The GNU Hurd is distributed in the hope that it will be useful, 
+   The GNU Hurd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
@@ -19,7 +19,6 @@
    the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 #include "priv.h"
-#include "fs_S.h"
 
 error_t
 trivfs_S_file_reparent (struct trivfs_protid *cred,
@@ -27,5 +26,9 @@ trivfs_S_file_reparent (struct trivfs_protid *cred,
 			mach_port_t parent,
 			mach_port_t *new, mach_msg_type_name_t *new_type)
 {
-  return EOPNOTSUPP;
+  /* This is not a directory, so we just duplicate it */
+  error_t ret = trivfs_S_io_duplicate (cred, reply, reply_type, new, new_type);
+  if (!ret)
+    mach_port_deallocate (mach_task_self (), parent);
+  return ret;
 }
