@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 1993, 1994, 1995, 1997 Free Software Foundation
+   Copyright (C) 1993,94,95,97,99,2002 Free Software Foundation, Inc.
 
 This file is part of the GNU Hurd.
 
@@ -8,7 +8,7 @@ it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.
 
-The GNU Hurd is distributed in the hope that it will be useful, 
+The GNU Hurd is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
@@ -20,7 +20,6 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 /* Written by Michael I. Bushnell.  */
 
 #include "priv.h"
-#include "fsys_S.h"
 #include "fsys_reply_U.h"
 #include <assert.h>
 #include <fcntl.h>
@@ -31,8 +30,8 @@ trivfs_S_fsys_getroot (struct trivfs_control *cntl,
 		       mach_port_t reply_port,
 		       mach_msg_type_name_t reply_port_type,
 		       mach_port_t dotdot,
-		       uid_t *uids, u_int nuids,
-		       uid_t *gids, u_int ngids,
+		       uid_t *uids, size_t nuids,
+		       uid_t *gids, size_t ngids,
 		       int flags,
 		       retry_type *do_retry,
 		       char *retry_name,
@@ -44,7 +43,6 @@ trivfs_S_fsys_getroot (struct trivfs_control *cntl,
   mach_port_t new_realnode;
   struct trivfs_protid *cred;
   struct iouser *user;
-  struct idvec *uvec, *gvec;
 
   if (!cntl)
     return EOPNOTSUPP;
@@ -58,12 +56,14 @@ trivfs_S_fsys_getroot (struct trivfs_control *cntl,
 	return err;
     }
 
+  if ((flags & O_WRITE & trivfs_allow_open) != (flags & O_WRITE))
+    return EROFS;
   if ((flags & (O_READ|O_WRITE|O_EXEC) & trivfs_allow_open)
       != (flags & (O_READ|O_WRITE|O_EXEC)))
-    return EOPNOTSUPP;
+    return EACCES;
 
   /* O_CREAT and O_EXCL are not meaningful here; O_NOLINK and O_NOTRANS
-     will only be useful when trivfs supports translators (which it doesn't 
+     will only be useful when trivfs supports translators (which it doesn't
      now). */
   flags &= O_HURD;
   flags &= ~(O_CREAT|O_EXCL|O_NOLINK|O_NOTRANS);
@@ -73,21 +73,33 @@ trivfs_S_fsys_getroot (struct trivfs_control *cntl,
   if (err)
     return err;
 
-  file_check_access (new_realnode, &perms);
+  err = iohelp_create_complex_iouser (&user, uids, nuids, gids, ngids);
+  if (err)
+    return err;
+
+  /* Validate permissions.  */
+  if (! trivfs_check_access_hook)
+    file_check_access (new_realnode, &perms);
+  else
+    (*trivfs_check_access_hook) (cntl, user, new_realnode, &perms);
   if ((flags & (O_READ|O_WRITE|O_EXEC) & perms)
       != (flags & (O_READ|O_WRITE|O_EXEC)))
     err = EACCES;
 
-  uvec = make_idvec ();
-  gvec = make_idvec ();
-  idvec_set_ids (uvec, uids, nuids);
-  idvec_set_ids (gvec, gids, ngids);
-  user = iohelp_create_iouser (uvec, gvec);
-  
   if (!err && trivfs_check_open_hook)
     err = (*trivfs_check_open_hook) (cntl, user, flags);
   if (!err)
-    err = trivfs_open (cntl, user, flags, new_realnode, &cred);
+    {
+      if (! trivfs_open_hook)
+	{
+	  err = trivfs_open (cntl, user, flags, new_realnode, &cred);
+	  if (!err)
+	    mach_port_deallocate (mach_task_self (), dotdot);
+	}
+      else
+	err = (*trivfs_open_hook) (cntl, user, dotdot, flags, new_realnode,
+				   &cred);
+    }
 
   if (err)
     {
@@ -101,7 +113,6 @@ trivfs_S_fsys_getroot (struct trivfs_control *cntl,
       *newpt = ports_get_right (cred);
       *newpttype = MACH_MSG_TYPE_MAKE_SEND;
       ports_port_deref (cred);
-      mach_port_deallocate (mach_task_self (), dotdot);
     }
 
   return err;
