@@ -1,5 +1,5 @@
 /* 
-   Copyright (C) 1996, 1997 Free Software Foundation, Inc.
+   Copyright (C) 1996, 1997, 2002 Free Software Foundation, Inc.
    Written by Michael I. Bushnell, p/BSG.
 
    This file is part of the GNU Hurd.
@@ -29,23 +29,34 @@
 #include <maptime.h>
 #include <argp.h>
 #include <argz.h>
+#include <error.h>
+#include <version.h>
+
+char *netfs_server_name = "nfs";
+char *netfs_server_version = HURD_VERSION;
 
 extern char *localhost ();
 
 /* Default number of times to retry RPCs when mounted soft. */
-#define DEFAULT_SOFT_RETRIES  3	
+#define DEFAULT_SOFT_RETRIES  3
 
 /* Default number of seconds to timeout cached stat information. */
 #define DEFAULT_STAT_TIMEOUT  3
 
 /* Default number of seconds to timeout cached file contents. */
-#define DEFAULT_CACHE_TIMEOUT 3	
+#define DEFAULT_CACHE_TIMEOUT 3
+
+/* Default number of seconds to timeout cache positive dir hits. */
+#define DEFAULT_NAME_CACHE_TIMEOUT 3
+
+/* Default number of seconds to timeout cache negative dir hits. */
+#define DEFAULT_NAME_CACHE_NEG_TIMEOUT 3
 
 /* Default maximum number of bytes to read at once. */
-#define DEFAULT_READ_SIZE     8192 
+#define DEFAULT_READ_SIZE     8192
 
 /* Default maximum number of bytes to write at once. */
-#define DEFAULT_WRITE_SIZE    8192 
+#define DEFAULT_WRITE_SIZE    8192
 
 
 /* Number of seconds to timeout cached stat information. */
@@ -54,6 +65,12 @@ int stat_timeout = DEFAULT_STAT_TIMEOUT;
 /* Number of seconds to timeout cached file contents. */
 int cache_timeout = DEFAULT_CACHE_TIMEOUT;
 
+/* Number of seconds to timeout cached positive dir hits. */
+int name_cache_timeout = DEFAULT_NAME_CACHE_TIMEOUT;
+
+/* Number of seconds to timeout cached negative dir hits. */
+int name_cache_neg_timeout = DEFAULT_NAME_CACHE_NEG_TIMEOUT;
+
 /* Number of seconds to wait for first retransmission of an RPC. */
 int initial_transmit_timeout = 1;
 
@@ -88,19 +105,23 @@ int write_size = DEFAULT_WRITE_SIZE;
 #define OPT_MNT_PROG    -11
 #define OPT_NFS_PROG    -12
 #define OPT_PMAP_PORT	-13
+#define OPT_NCACHE_TO	-14
+#define OPT_NCACHE_NEG_TO -15
 
 /* Return a string corresponding to the printed rep of DEFAULT_what */
 #define ___D(what) #what
 #define __D(what) ___D(what)
 #define _D(what) __D(DEFAULT_ ## what)
 
+const char *argp_program_version = STANDARD_HURD_VERSION (nfs);
+
 /* Options usable both at startup and at runtime.  */
 static const struct argp_option common_options[] =
 {
   {0,0,0,0,0,1},
   {"soft",		    OPT_SOFT,	   "RETRIES", OPTION_ARG_OPTIONAL,
-     "File system requests will eventually fail, after RETRIES tries if"
-     " specified, otherwise " _D(SOFT_RETRIES)},
+     "File system requests will eventually fail, after RETRIES tries"
+     " (default " _D(SOFT_RETRIES) ")" },
   {"hard",		    OPT_HARD, 0, 0,
      "Retry file systems requests until they succeed"},
 
@@ -117,6 +138,12 @@ static const struct argp_option common_options[] =
      "Timeout for cached stat information (default " _D(STAT_TIMEOUT) ")"},
   {"cache-timeout",	    OPT_CACHE_TO,  "SEC", 0,
      "Timeout for cached file data (default " _D(CACHE_TIMEOUT) ")"},
+  {"name-cache-timeout",    OPT_NCACHE_TO, "SEC", 0,
+     "Timeout for positive directory cache entries (default " 
+     _D(NAME_CACHE_TIMEOUT) ")"},
+  {"name-cache-neg-timeout", OPT_NCACHE_NEG_TO, "SEC", 0,
+     "Timeout for negative directory cache entires (default "
+      _D(NAME_CACHE_NEG_TIMEOUT) ")"},
   {"init-transmit-timeout", OPT_INIT_TR_TO,"SEC", 0}, 
   {"max-transmit-timeout",  OPT_MAX_TR_TO, "SEC", 0}, 
 
@@ -144,6 +171,8 @@ parse_common_opt (int key, char *arg, struct argp_state *state)
     case OPT_CACHE_TO: cache_timeout = atoi (arg); break;
     case OPT_INIT_TR_TO: initial_transmit_timeout = atoi (arg); break;
     case OPT_MAX_TR_TO: max_transmit_timeout = atoi (arg); break;
+    case OPT_NCACHE_TO: name_cache_timeout = atoi (arg); break;
+    case OPT_NCACHE_NEG_TO: name_cache_neg_timeout = atoi (arg); break;
 
     default:
       return ARGP_ERR_UNKNOWN;
@@ -174,7 +203,7 @@ static const struct argp_option startup_options[] = {
 static char *args_doc = "REMOTE_FS [HOST]";
 static char *doc = "Hurd nfs translator"
 "\vIf HOST is not specified, an attempt is made to extract"
-" it from REMOTE_FS, using either the `HOST:FS' or `FS@HOST' notations.";
+" it from REMOTE_FS using either the `HOST:FS' or `FS@HOST' notations.";
 
 static const struct argp_child
 runtime_argp_children[] = { {&netfs_std_runtime_argp}, {0} };
@@ -182,12 +211,12 @@ static struct argp
 runtime_argp = { common_options, parse_common_opt, 0, 0,
 		 runtime_argp_children };
 
-/* Use by netfs_set_options to handle runtime option parsing.  */
+/* Used by netfs_set_options to handle runtime option parsing.  */
 struct argp *netfs_runtime_argp = &runtime_argp;
 
 /* Where to find the remote filesystem.  */
-static char *remote_fs = 0;
-static char *host = 0;
+static char *remote_fs; /* = 0; */
+static char *host; /* = 0; */
 
 /* Return an argz string describing the current options.  Fill *ARGZ
    with a pointer to newly malloced storage holding the list and *LEN
@@ -219,6 +248,8 @@ netfs_append_args (char **argz, size_t *argz_len)
   FOPT ("--cache-timeout=%d", cache_timeout);
   FOPT ("--init-transmit-timeout=%d", initial_transmit_timeout);
   FOPT ("--max-transmit-timeout=%d", max_transmit_timeout);
+  FOPT ("--name-cache-timeout=%d", name_cache_timeout);
+  FOPT ("--name-cache-neg-timeout=%d", name_cache_neg_timeout);
 
   if (! err)
     err = netfs_append_std_options (argz, argz_len);
@@ -247,6 +278,8 @@ extract_nfs_args (char *spec, char **remote_fs, char **host)
   char *sep;
 
   spec = strdup (spec);		/* So we can trash it.  */
+  if (! spec)
+    return NULL;
 
   sep = index (spec, ':');
   if (sep)
@@ -318,6 +351,9 @@ parse_startup_opt (int key, char *arg, struct argp_state *state)
 int
 main (int argc, char **argv)
 {
+  pthread_t thread;
+  error_t err;
+
   struct argp common_argp = { common_options, parse_common_opt };
   const struct argp_child argp_children[] =
     { {&common_argp}, {&netfs_std_startup_argp}, {0} };
@@ -351,17 +387,28 @@ main (int argc, char **argv)
     }
   while ((ret == -1) && (errno == EADDRINUSE));
   if (ret == -1)
+    error (1, errno, "binding main udp socket");
+
+  err = maptime_map (0, 0, &mapped_time);
+  if (err)
+    error (2, err, "mapping time");
+
+  err = pthread_create (&thread, NULL, timeout_service_thread, NULL);
+  if (!err)
+    pthread_detach (thread);
+  else
     {
-      perror ("binding main udp socket");
-      exit (1);
+      errno = err;
+      perror ("pthread_create");
+    }
+  err = pthread_create (&thread, NULL, rpc_receive_thread, NULL);
+  if (!err)
+    pthread_detach (thread);
+  else
+    {
+      errno = err;
+      perror ("pthread_create");
     }
-
-  errno = maptime_map (0, 0, &mapped_time);
-  if (errno)
-    perror ("mapping time");
-
-  cthread_detach (cthread_fork ((cthread_fn_t) timeout_service_thread, 0));
-  cthread_detach (cthread_fork ((cthread_fn_t) rpc_receive_thread, 0));
   
   hostname = localhost ();
 
