@@ -1,8 +1,7 @@
 /* Print vm statistics
 
-   Copyright (C) 1996, 1997 Free Software Foundation, Inc.
-
-   Written by Miles Bader <miles@gnu.ai.mit.edu>
+   Copyright (C) 1996,97,99,2002 Free Software Foundation, Inc.
+   Written by Miles Bader <miles@gnu.org>
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -25,12 +24,16 @@
 #include <string.h>
 #include <unistd.h>
 #include <stdlib.h>
+#include <fcntl.h>
 #include <version.h>
 
 #include <mach.h>
+#include <mach/gnumach.h>
 #include <mach/vm_statistics.h>
+#include <mach/vm_cache_statistics.h>
 #include <mach/default_pager.h>
 #include <hurd.h>
+#include <hurd/paths.h>
 
 const char *argp_program_version = STANDARD_HURD_VERSION (vmstat);
 
@@ -59,6 +62,7 @@ static const char *doc = "Show system virtual memory statistics"
    than what the system returns values in, as we represent some quantities as
    bytes instead of pages)!  */
 typedef long long val_t;
+#define BADVAL ((val_t) -1LL)	/* a good generic value for "couldn't get" */
 
 /* What a given number describes.  */
 enum val_type
@@ -103,13 +107,13 @@ print_val (val_t val, enum val_type type,
 	   size_t size_units, int fwidth, int sign)
 {
   if (type == PCENT)
-    printf (sign ? "%+*d%%" : "%*d%%", fwidth - 1, val);
+    printf (sign ? "%+*lld%%" : "%*lld%%", fwidth - 1, val);
   else if ((type == SIZE || type == PAGESZ) && size_units == 0)
     {
       float fval = val;
       char *units = " KMGT", *u = units;
 
-      while (fval > 1024)
+      while (fval >= 10000)
 	{
 	  fval /= 1024;
 	  u++;
@@ -124,7 +128,7 @@ print_val (val_t val, enum val_type type,
     {
       if ((type == SIZE || type == PAGESZ) && size_units > 0)
 	val /= size_units;
-      printf (sign ? "%+*d" : "%*d", fwidth, val);
+      printf (sign ? "%+*lld" : "%*lld", fwidth, val);
     }
 }
 
@@ -178,11 +182,14 @@ struct field
 };
 
 /* State about system vm from which we compute the above defined fields.  */
-struct vm_state 
+struct vm_state
 {
   /* General vm statistics.  */
   struct vm_statistics vmstats;
 
+  /* Page cache statistics.  */
+  struct vm_cache_statistics cache_stats;
+
   /* default pager port (must be privileged to fetch this).  */
   mach_port_t def_pager;
   struct default_pager_info def_pager_info;
@@ -193,11 +200,16 @@ vm_state_refresh (struct vm_state *state)
 {
   error_t err = vm_statistics (mach_task_self (), &state->vmstats);
 
+  if (err)
+    return err;
+
+  err = vm_cache_statistics (mach_task_self (), &state->cache_stats);
+
   if (err)
     return err;
 
   /* Mark the info as invalid, but leave DEF_PAGER alone.  */
-  bzero (&state->def_pager_info, sizeof state->def_pager_info);
+  memset (&state->def_pager_info, 0, sizeof state->def_pager_info);
 
   return 0;
 }
@@ -245,62 +257,64 @@ ensure_def_pager_info (struct vm_state *state)
       mach_port_t host;
 
       err = get_privileged_ports (&host, 0);
-      if (err)
+      if (err == EPERM)
 	{
-	  error (0, err, "get_privileged_ports");
-	  return 0;
+	  /* We are not root, so try opening the /servers file.  */
+	  state->def_pager = file_name_lookup (_SERVERS_DEFPAGER, O_READ, 0);
+	  if (state->def_pager == MACH_PORT_NULL)
+	    {
+	      error (0, errno, _SERVERS_DEFPAGER);
+	      return 0;
+	    }
 	}
-
-      err = vm_set_default_memory_manager (host, &state->def_pager);
-      mach_port_deallocate (mach_task_self (), host);
-
-      if (err)
+      if (state->def_pager == MACH_PORT_NULL)
 	{
-	  error (0, err, "vm_set_default_memory_manager");
-	  return 0;
+	  if (err)
+	    {
+	      error (0, err, "get_privileged_ports");
+	      return 0;
+	    }
+
+	  err = vm_set_default_memory_manager (host, &state->def_pager);
+	  mach_port_deallocate (mach_task_self (), host);
+
+	  if (err)
+	    {
+	      error (0, err, "vm_set_default_memory_manager");
+	      return 0;
+	    }
 	}
     }
 
+  if (!MACH_PORT_VALID (state->def_pager))
+    {
+      if (state->def_pager == MACH_PORT_NULL)
+	{
+	  error (0, 0,
+		 "No default pager running, so no swap information available");
+	  state->def_pager = MACH_PORT_DEAD; /* so we don't try again */
+	}
+      return 0;
+    }
+
   err = default_pager_info (state->def_pager, &state->def_pager_info);
   if (err)
     error (0, err, "default_pager_info");
-
   return (err == 0);
 }
 
-static val_t
-get_swap_size (struct vm_state *state, const struct field *field)
-{
-  return
-    ensure_def_pager_info (state) ? state->def_pager_info.dpi_total_space : -1;
-}
+#define SWAP_FIELD(getter, expr) \
+  static val_t getter (struct vm_state *state, const struct field *field) \
+  { return ensure_def_pager_info (state) ? (val_t) (expr) : BADVAL; }
 
-static val_t
-get_swap_free (struct vm_state *state, const struct field *field)
-{
-  return
-    ensure_def_pager_info (state) ? state->def_pager_info.dpi_free_space : -1;
-}
-
-static val_t
-get_swap_page_size (struct vm_state *state, const struct field *field)
-{
-  return
-    ensure_def_pager_info (state) ? state->def_pager_info.dpi_page_size : -1;
-}
-
-static val_t
-get_swap_active (struct vm_state *state, const struct field *field)
-{
-  return
-    ensure_def_pager_info (state)
-    ? (state->def_pager_info.dpi_total_space
-       - state->def_pager_info.dpi_free_space)
-    : -1;
-}
+SWAP_FIELD (get_swap_size, state->def_pager_info.dpi_total_space)
+SWAP_FIELD (get_swap_free, state->def_pager_info.dpi_free_space)
+SWAP_FIELD (get_swap_page_size, state->def_pager_info.dpi_page_size)
+SWAP_FIELD (get_swap_active, (state->def_pager_info.dpi_total_space
+			      - state->def_pager_info.dpi_free_space))
 
-/* Returns the byte offset of the field FIELD in a vm_statistics structure. */
-#define _F(field_name)  offsetof (struct vm_statistics, field_name)
+/* Returns the byte offset of the field FIELD in a vm_state structure. */
+#define _F(field_name)  offsetof (struct vm_state, field_name)
 
 #define K 1024
 #define M (1024*K)
@@ -310,35 +324,39 @@ get_swap_active (struct vm_state *state, const struct field *field)
 static const struct field fields[] =
 {
   {"pagesize",	   "pgsz", "System pagesize",
-   CONST, PAGESZ, 16*K,		1, _F (pagesize) },
+   CONST, PAGESZ, 16*K,		1, _F (vmstats.pagesize) },
   {"size",	   "size", "Usable physical memory",
    CONST, SIZE,   VAL_MAX_MEM,	1, 0, get_size },
   {"free",	   "free", "Unused physical memory",
-   VARY,  SIZE,   VAL_MAX_MEM,	1, _F (free_count) },
+   VARY,  SIZE,   VAL_MAX_MEM,	1, _F (vmstats.free_count) },
   {"active",	   "actv", "Physical memory in active use",
-   VARY,  SIZE,   VAL_MAX_MEM,	1, _F (active_count) },
+   VARY,  SIZE,   VAL_MAX_MEM,	1, _F (vmstats.active_count) },
   {"inactive", 	   "inact", "Physical memory in the inactive queue",
-   VARY,  SIZE,   VAL_MAX_MEM,	1, _F (inactive_count) },
+   VARY,  SIZE,   VAL_MAX_MEM,	1, _F (vmstats.inactive_count) },
   {"wired",    	   "wired", "Unpageable physical memory",
-   VARY,  SIZE,   VAL_MAX_MEM,	1, _F (wire_count) },
+   VARY,  SIZE,   VAL_MAX_MEM,	1, _F (vmstats.wire_count) },
   {"zero filled",  "zeroed","Cumulative zero-filled pages",
-   CUMUL, SIZE,   90*G,		1, _F (zero_fill_count) },
+   CUMUL, SIZE,   90*G,		1, _F (vmstats.zero_fill_count) },
   {"reactivated",  "react", "Cumulative reactivated inactive pages",
-   CUMUL, SIZE,   900*M,	1, _F (reactivations) },
+   CUMUL, SIZE,   900*M,	1, _F (vmstats.reactivations) },
   {"pageins",	   "pgins", "Cumulative pages paged in",
-   CUMUL, SIZE,   90*G,		1, _F (pageins) },
+   CUMUL, SIZE,   90*G,		1, _F (vmstats.pageins) },
   {"pageouts",	   "pgouts","Cumulative pages paged out",
-   CUMUL, SIZE,   90*G,		1, _F (pageouts) },
+   CUMUL, SIZE,   90*G,		1, _F (vmstats.pageouts) },
   {"page faults",  "pfaults","Cumulative page faults",
-   CUMUL, COUNT,  99999999,	1, _F (faults) },
+   CUMUL, COUNT,  99999999,	1, _F (vmstats.faults) },
   {"cow faults",   "cowpfs", "Cumulative copy-on-write page faults",
-   CUMUL, COUNT,  9999999,	1, _F (cow_faults) },
+   CUMUL, COUNT,  9999999,	1, _F (vmstats.cow_faults) },
   {"memobj lookups","lkups","Memory-object lookups",
-   CUMUL, COUNT,  999999,	0, _F (lookups) },
+   CUMUL, COUNT,  999999,	0, _F (vmstats.lookups) },
   {"memobj hits",   "hits", "Memory-object lookups with active pagers",
-   CUMUL, COUNT,  999999,	0, _F (hits) },
+   CUMUL, COUNT,  999999,	0, _F (vmstats.hits) },
   {"memobj hit ratio","hrat","Percentage of memory-object lookups with active pagers",
    VARY, PCENT,   99,		1, -1, get_memobj_hit_ratio },
+  {"cached memobjs", "caobj", "Number of memory-objects retained in the page cache",
+   VARY,  COUNT,  99999999,     1, _F (cache_stats.cache_object_count) },
+  {"cache",        "cache", "Physical memory used by the page cache",
+   VARY,  SIZE,   VAL_MAX_MEM,  1, _F (cache_stats.cache_count) },
   {"swap size",	   "swsize", "Size of the default-pager swap area",
    CONST, SIZE,   VAL_MAX_SWAP,	1, 0 ,get_swap_size },
   {"swap active",  "swactv", "Default-pager swap area in use",
@@ -427,7 +445,7 @@ main (int argc, char **argv)
   /* Construct an options vector for them.  */
   field_opts_size = ((num_fields + 1) * sizeof (struct argp_option));
   field_opts = alloca (field_opts_size);
-  bzero (field_opts, field_opts_size);
+  memset (field_opts, 0, field_opts_size);
 
   for (field = fields; field->name; field++)
     {
@@ -439,7 +457,7 @@ main (int argc, char **argv)
       opt->doc = field->doc;
       opt->group = 2;
     }
-  /* No need to terminate FIELD_OPTS because the bzero above's done so.  */
+  /* No need to terminate FIELD_OPTS because the memset above has done so.  */
 
   field_argp.options = field_opts;
 
@@ -452,7 +470,7 @@ main (int argc, char **argv)
       if (field->standard)
 	output_fields |= (1 << (field - fields));
 
-  /* Returns an appropiate SIZE_UNITS for printing FIELD.  */
+  /* Returns an appropriate SIZE_UNITS for printing FIELD.  */
 #define SIZE_UNITS(field)						      \
   (size_units >= 0							      \
    ? size_units								      \
@@ -472,7 +490,7 @@ main (int argc, char **argv)
 	       (field)->type, SIZE_UNITS (field))
 
   /* Actually fetch the statistics.  */
-  bzero (&state, sizeof (state)); /* Initialize STATE.  */
+  memset (&state, 0, sizeof (state)); /* Initialize STATE.  */
   err = vm_state_refresh (&state);
   if (err)
     error (2, err, "vm_state_refresh");
@@ -501,7 +519,7 @@ main (int argc, char **argv)
       do
 	{
 	  int num;
-	  size_t fwidths[num_fields];
+	  int fwidths[num_fields];
 
 	  if (first_hdr)
 	    first_hdr = 0;
@@ -556,7 +574,7 @@ main (int argc, char **argv)
 		  }
 	      putchar ('\n');
 	    }
-	
+
 	  prev_state = state;
 
 	  for (repeats = 0
@@ -567,17 +585,24 @@ main (int argc, char **argv)
 	      for (field = fields, num = 0, first = 1; field->name; field++, num++)
 		if (output_fields & (1 << (field - fields)))
 		  {
-		    int sign = 0;
 		    val_t val = vm_state_get_field (&state, field);
 
-		    if (repeats && field->change_type == CUMUL)
+		    if (val < 0)
+		      /* Couldn't fetch this field, don't try again.  */
+		      const_fields &= ~(1 << (field - fields));
+		    else
 		      {
-			sign = 1;
-			val -= vm_state_get_field (&prev_state, field);
-		      }
+			int sign = 0;
 
-		    PSEP (" ", 0);
-		    PVAL (val, field, fwidths[num], sign);
+			if (repeats && field->change_type == CUMUL)
+			  {
+			    sign = 1;
+			    val -= vm_state_get_field (&prev_state, field);
+			  }
+
+			PSEP (" ", 0);
+			PVAL (val, field, fwidths[num], sign);
+		      }
 		  }
 	      putchar ('\n');
 
@@ -617,14 +642,17 @@ main (int argc, char **argv)
 	if (output_fields & (1 << (field - fields)))
 	  {
 	    val_t val = vm_state_get_field (&state, field);
-	    int fwidth = 0;
-	    if (print_prefix)
+	    if (val >= 0)
 	      {
-		printf ("%s: ", field->name);
-		fwidth = max_width - strlen (field->name);
+		int fwidth = 0;
+		if (print_prefix)
+		  {
+		    printf ("%s: ", field->name);
+		    fwidth = max_width - strlen (field->name);
+		  }
+		PVAL (val, field, fwidth, 0);
+		putchar ('\n');
 	      }
-	    PVAL (val, field, fwidth, 0);
-	    putchar ('\n');
 	  }
     }
 
